@@ -1,6 +1,8 @@
 import inspect
+import itertools
 from collections import Iterable
 import re
+import threading
 
 import numpy as np
 import scipy
@@ -10,27 +12,25 @@ from Orange.data.util import one_hot
 from Orange.misc.wrapper_meta import WrapperMeta
 from Orange.preprocess import (RemoveNaNClasses, Continuize,
                                RemoveNaNColumns, SklImpute, Normalize)
-from Orange.util import Reprable, patch
+from Orange.util import Reprable
 
 __all__ = ["Learner", "Model", "SklLearner", "SklModel"]
 
 
 class _ReprableWithPreprocessors(Reprable):
-    def __repr__(self):
-        preprocessors = self.preprocessors
-        # TODO: Implement __eq__ on preprocessors to avoid comparing reprs
-        if repr(preprocessors) == repr(list(self.__class__.preprocessors)):
-            # If they are default, set to None temporarily to avoid printing
-            preprocessors = None
-        with patch.object(self, 'preprocessors', preprocessors):
-            return super().__repr__()
-
-
-class _ReprableWithParams(Reprable):
-    def __repr__(self):
-        # In addition to saving values onto self, SklLearners save into params
-        with patch.object(self, '__dict__', dict(self.__dict__, **self.params)):
-            return super().__repr__()
+    def _reprable_omit_param(self, name, default, value):
+        if name == "preprocessors":
+            default_cls = type(self).preprocessors
+            if value is default or value is default_cls:
+                return True
+            else:
+                try:
+                    return all(p1 is p2 for p1, p2 in
+                               itertools.zip_longest(value, default_cls))
+                except (ValueError, TypeError):
+                    return False
+        else:
+            return super()._reprable_omit_param(name, default, value)
 
 
 class Learner(_ReprableWithPreprocessors):
@@ -88,6 +88,7 @@ class Learner(_ReprableWithPreprocessors):
             self.preprocessors = tuple(preprocessors)
         elif preprocessors:
             self.preprocessors = (preprocessors,)
+        self.__tls = threading.local()
 
     def fit(self, X, Y, W=None):
         raise RuntimeError(
@@ -169,6 +170,31 @@ class Learner(_ReprableWithPreprocessors):
     def name(self, value):
         self.__name = value
 
+    # Learners implemented using `fit` access the `domain` through the
+    # instance attribute. This makes (or it would) make it impossible to
+    # be implemented in a thread-safe manner. So the domain is made a
+    # property descriptor utilizing thread local storage behind the scenes.
+    @property
+    def domain(self):
+        return self.__tls.domain
+
+    @domain.setter
+    def domain(self, domain):
+        self.__tls.domain = domain
+
+    @domain.deleter
+    def domain(self):
+        del self.__tls.domain
+
+    def __getstate__(self):
+        state = dict(self.__dict__)
+        del state["_Learner__tls"]
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        self.__tls = threading.local()
+
     def __str__(self):
         return self.name
 
@@ -222,13 +248,13 @@ class Model(Reprable):
             prediction = self.predict_storage(data)
         elif isinstance(data, Table):
             if data.domain != self.domain:
-                data = data.from_table(self.domain, data)
+                data = data.transform(self.domain)
             prediction = self.predict_storage(data)
         elif isinstance(data, (list, tuple)):
             if not isinstance(data[0], (list, tuple)):
                 data = [data]
             data = Table(self.original_domain, data)
-            data = Table(self.domain, data)
+            data = data.transform(self.domain)
             prediction = self.predict_storage(data)
         else:
             raise TypeError("Unrecognized argument (instance of '{}')"
@@ -273,6 +299,14 @@ class Model(Reprable):
         else:  # ret == Model.ValueProbs
             return value, probs
 
+    def __getstate__(self):
+        """Skip (possibly large) data when pickling models"""
+        state = self.__dict__
+        if 'original_data' in state:
+            state = state.copy()
+            state['original_data'] = None
+        return state
+
 
 class SklModel(Model, metaclass=WrapperMeta):
     used_vals = None
@@ -292,7 +326,7 @@ class SklModel(Model, metaclass=WrapperMeta):
         return super().__repr__() + '  # params=' + repr(self.params)
 
 
-class SklLearner(_ReprableWithParams, Learner, metaclass=WrapperMeta):
+class SklLearner(Learner, metaclass=WrapperMeta):
     """
     ${skldoc}
     Additional Orange parameters
@@ -360,6 +394,18 @@ class SklLearner(_ReprableWithParams, Learner, metaclass=WrapperMeta):
         """
         return 'sample_weight' in self.__wraps__.fit.__code__.co_varnames
 
+    def __getattr__(self, item):
+        try:
+            return self.params[item]
+        except (KeyError, AttributeError):
+            raise AttributeError(item) from None
+
+    # TODO: Disallow (or mirror) __setattr__ for keys in params?
+
+    def __dir__(self):
+        dd = super().__dir__()
+        return list(sorted(set(dd) | set(self.params.keys())))
+
 
 class TreeModel(Model):
     pass
