@@ -32,6 +32,7 @@ so they can be used alter. It should be called before widget starts modifying
 import copy
 import itertools
 import os
+import logging
 import pickle
 import time
 import warnings
@@ -40,6 +41,8 @@ from Orange.data import Domain, Variable
 from Orange.misc.environ import widget_settings_dir
 from Orange.widgets.utils import vartype
 
+log = logging.getLogger(__name__)
+
 __all__ = ["Setting", "SettingsHandler",
            "ContextSetting", "ContextHandler",
            "DomainContextHandler", "PerfectDomainContextHandler",
@@ -397,15 +400,20 @@ class SettingsHandler:
         should overload the latter."""
         filename = self._get_settings_filename()
         os.makedirs(os.path.dirname(filename), exist_ok=True)
-
-        settings_file = open(filename, "wb")
         try:
-            self.write_defaults_file(settings_file)
-        except (EOFError, IOError, pickle.PicklingError):
-            settings_file.close()
-            os.remove(filename)
-        else:
-            settings_file.close()
+            settings_file = open(filename, "wb")
+            try:
+                self.write_defaults_file(settings_file)
+            except (EOFError, IOError, pickle.PicklingError) as ex:
+                log.error("Could not write default settings for %s (%s).",
+                          self.widget_class, type(ex).__name__)
+                settings_file.close()
+                os.remove(filename)
+            else:
+                settings_file.close()
+        except PermissionError as ex:
+            log.error("Could not write default settings for %s (%s).",
+                      self.widget_class, type(ex).__name__)
 
     def write_defaults_file(self, settings_file):
         """Write defaults for this widget class to a file
