@@ -1,5 +1,6 @@
 import time
 from enum import IntEnum
+from collections import OrderedDict
 
 import numpy as np
 
@@ -12,6 +13,7 @@ from AnyQt.QtGui import QColor, QPainter, QFont, QPen, QBrush
 from AnyQt.QtCore import Qt, QEvent, QRectF, QSize
 
 from Orange.data import Table, Domain
+from Orange.statistics.util import nanmin, nanmax, mean, unique
 from Orange.classification import Model
 from Orange.classification.naive_bayes import NaiveBayesModel
 from Orange.classification.logistic_regression import \
@@ -652,6 +654,7 @@ class OWNomogram(OWWidget):
         self.hidden_vertical_line = None
         self.old_target_class_index = self.target_class_index
         self.markers_set = False
+        self.repaint = False
 
         # GUI
         box = gui.vBox(self.controlArea, "Target class")
@@ -765,6 +768,7 @@ class OWNomogram(OWWidget):
 
     def eventFilter(self, obj, event):
         if obj is self.view.viewport() and event.type() == QEvent.Resize:
+            self.repaint = True
             values = [item.dot.value for item in self.feature_items]
             self.feature_marker_values = self.scale_back(values)
             self.update_scene()
@@ -850,7 +854,7 @@ class OWNomogram(OWWidget):
 
         self.domain = self.reconstruct_domain(self.classifier.original_domain,
                                               self.domain)
-        self.data = Table.from_table(self.domain, self.classifier.original_data)
+        self.data = self.classifier.original_data.transform(self.domain)
         attrs, ranges, start = self.domain.attributes, [], 0
         for attr in attrs:
             stop = start + len(attr.values) if attr.is_discrete else start + 1
@@ -865,11 +869,13 @@ class OWNomogram(OWWidget):
         self.log_reg_coeffs = [coeffs[:, ranges[i]] for i in range(len(attrs))]
         self.log_reg_coeffs_orig = self.log_reg_coeffs.copy()
 
-        for i in range(len(self.log_reg_coeffs)):
+        min_values = nanmin(self.data.X, axis=0)
+        max_values = nanmax(self.data.X, axis=0)
+
+        for i, min_t, max_t in zip(range(len(self.log_reg_coeffs)),
+                                   min_values, max_values):
             if self.log_reg_coeffs[i].shape[1] == 1:
                 coef = self.log_reg_coeffs[i]
-                min_t = np.nanmin(self.data.X, axis=0)[i]
-                max_t = np.nanmax(self.data.X, axis=0)[i]
                 self.log_reg_coeffs[i] = np.hstack((coef * min_t, coef * max_t))
                 self.log_reg_cont_data_extremes.append(
                     [sorted([min_t, max_t], reverse=(c < 0)) for c in coef])
@@ -877,6 +883,8 @@ class OWNomogram(OWWidget):
                 self.log_reg_cont_data_extremes.append([None])
 
     def update_scene(self):
+        if not self.repaint:
+            return
         self.clear_scene()
         if self.domain is None or not len(self.points[0]):
             return
@@ -1076,10 +1084,10 @@ class OWNomogram(OWWidget):
             value, feature_val = 0, None
             if len(self.log_reg_coeffs):
                 if attr.is_discrete:
-                    ind, n = np.unique(self.data.X[:, i], return_counts=True)
+                    ind, n = unique(self.data.X[:, i], return_counts=True)
                     feature_val = np.nan_to_num(ind[np.argmax(n)])
                 else:
-                    feature_val = np.average(self.data.X[:, i])
+                    feature_val = mean(self.data.X[:, i])
             inst_in_dom = instances and attr in instances.domain
             if inst_in_dom and not np.isnan(instances[0][attr]):
                 feature_val = instances[0][attr]
@@ -1104,13 +1112,15 @@ class OWNomogram(OWWidget):
 
     @staticmethod
     def reconstruct_domain(original, preprocessed):
-        attrs = []
+        # abuse dict to make "in" comparisons faster
+        attrs = OrderedDict()
         for attr in preprocessed.attributes:
             cv = attr._compute_value.variable._compute_value
             var = cv.variable if cv else original[attr.name]
-            if var in attrs:
+            if var in attrs:    # the reason for OrderedDict
                 continue
-            attrs.append(var)
+            attrs[var] = None   # we only need keys
+        attrs = list(attrs.keys())
         return Domain(attrs, original.class_var, original.metas)
 
     @staticmethod
