@@ -32,7 +32,10 @@ from Orange.widgets.utils.annotated_data import (create_annotated_table,
 from Orange.widgets.utils.sql import check_sql_input
 from Orange.widgets.unsupervised.owhierarchicalclustering import \
     WrapperLayoutItem
-from Orange.widgets.widget import Msg
+from Orange.widgets.widget import Msg, Input, Output
+
+
+ROW_NAMES_WIDTH = 200
 
 
 class OWSilhouettePlot(widget.OWWidget):
@@ -43,9 +46,12 @@ class OWSilhouettePlot(widget.OWWidget):
     icon = "icons/SilhouettePlot.svg"
     priority = 300
 
-    inputs = [("Data", Orange.data.Table, "set_data")]
-    outputs = [("Selected Data", Orange.data.Table, widget.Default),
-               (ANNOTATED_DATA_SIGNAL_NAME, Orange.data.Table)]
+    class Inputs:
+        data = Input("Data", Orange.data.Table)
+
+    class Outputs:
+        selected_data = Output("Selected Data", Orange.data.Table, default=True)
+        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Orange.data.Table)
 
     replaces = [
         "orangecontrib.prototypes.widgets.owsilhouetteplot.OWSilhouettePlot",
@@ -66,7 +72,7 @@ class OWSilhouettePlot(widget.OWWidget):
     bar_size = settings.Setting(3)
     #: Add silhouette scores to output data
     add_scores = settings.Setting(False)
-    auto_commit = settings.Setting(False)
+    auto_commit = settings.Setting(True)
 
     Distances = [("Euclidean", Orange.distance.Euclidean),
                  ("Manhattan", Orange.distance.Manhattan)]
@@ -157,6 +163,7 @@ class OWSilhouettePlot(widget.OWWidget):
         return sh.expandedTo(QSize(600, 720))
 
     @check_sql_input
+    @Inputs.data
     def set_data(self, data):
         """
         Set the input data set.
@@ -246,7 +253,7 @@ class OWSilhouettePlot(widget.OWWidget):
 
     def _update(self):
         # Update/recompute the distances/scores as required
-        if self.data is None:
+        if self.data is None or not len(self.data):
             self._mask = None
             self._silhouette = None
             self._labels = None
@@ -260,6 +267,7 @@ class OWSilhouettePlot(widget.OWWidget):
 
         labelvar = self.cluster_var_model[self.cluster_var_idx]
         labels, _ = self.data.get_column_view(labelvar)
+        labels = numpy.asarray(labels, dtype=float)
         mask = numpy.isnan(labels)
         labels = labels.astype(int)
         labels = labels[~mask]
@@ -305,12 +313,13 @@ class OWSilhouettePlot(widget.OWWidget):
             self._set_bar_height()
 
             if self.group_by_cluster:
-                silplot.setScores(self._silhouette, self._labels, var.values)
+                silplot.setScores(self._silhouette, self._labels, var.values,
+                                  var.colors)
             else:
                 silplot.setScores(
                     self._silhouette,
                     numpy.zeros(len(self._silhouette), dtype=int),
-                    [""]
+                    [""], numpy.array([[63, 207, 207]])
                 )
 
             self.scene.addItem(silplot)
@@ -379,8 +388,7 @@ class OWSilhouettePlot(widget.OWWidget):
                     self.data.domain.attributes,
                     self.data.domain.class_vars,
                     self.data.domain.metas + (silhouette_var, ))
-                data = self.data.from_table(
-                    domain, self.data)
+                data = self.data.transform(domain)
             else:
                 domain = self.data.domain
                 data = self.data
@@ -394,9 +402,8 @@ class OWSilhouettePlot(widget.OWWidget):
                     selected[:, silhouette_var] = numpy.c_[scores[selectedmask]]
                 data[:, silhouette_var] = numpy.c_[scores]
 
-        self.send("Selected Data", selected)
-        self.send(ANNOTATED_DATA_SIGNAL_NAME,
-                  create_annotated_table(data, indices))
+        self.Outputs.selected_data.send(selected)
+        self.Outputs.annotated_data.send(create_annotated_table(data, indices))
 
     def send_report(self):
         if not len(self.cluster_var_model):
@@ -437,13 +444,13 @@ class SilhouettePlot(QGraphicsWidget):
         self.__selection = numpy.asarray([], dtype=int)
         self.__selstate = None
         self.__pen = QPen(Qt.NoPen)
-        self.__brush = QBrush(QColor("#3FCFCF"))
         self.__layout = QGraphicsGridLayout()
         self.__hoveredItem = None
         self.setLayout(self.__layout)
         self.layout().setColumnSpacing(0, 1.)
+        self.setFocusPolicy(Qt.StrongFocus)
 
-    def setScores(self, scores, labels, values, rownames=None):
+    def setScores(self, scores, labels, values, colors, rownames=None):
         """
         Set the silhouette scores/labels to for display.
 
@@ -455,6 +462,8 @@ class SilhouettePlot(QGraphicsWidget):
             A ndarray (dtype=int) of label/clusters indices.
         values : list of str
             A list of label/cluster names.
+        colors : (N, 3) ndarray
+            A ndarray of RGB values.
         rownames : list of str, optional
             A list (len == N) of row names.
         """
@@ -481,8 +490,9 @@ class SilhouettePlot(QGraphicsWidget):
         groups = [
             namespace(scores=scores[indices], indices=indices, label=label,
                       rownames=(rownames[indices] if rownames is not None
-                                else None))
-            for indices, label in zip(cluster_indices, values)
+                                else None),
+                      color=color)
+            for indices, label, color in zip(cluster_indices, values, colors)
         ]
         self.clear()
         self.__groups = groups
@@ -502,7 +512,10 @@ class SilhouettePlot(QGraphicsWidget):
             item = layout.itemAt(i + 1, 3)
 
             if grp.rownames is not None:
-                item.setItems(grp.rownames)
+                metrics = QFontMetrics(self.font())
+                rownames = [metrics.elidedText(rowname, Qt.ElideRight, ROW_NAMES_WIDTH)
+                            for rowname in grp.rownames]
+                item.setItems(rownames)
                 item.setVisible(self.__rowNamesVisible)
             else:
                 item.setItems([])
@@ -581,7 +594,7 @@ class SilhouettePlot(QGraphicsWidget):
 
         for i, group in enumerate(self.__groups):
             silhouettegroup = BarPlotItem(parent=self)
-            silhouettegroup.setBrush(self.__brush)
+            silhouettegroup.setBrush(QBrush(QColor(*group.color)))
             silhouettegroup.setPen(self.__pen)
             silhouettegroup.setDataRange(smin, smax)
             silhouettegroup.setPlotData(group.scores)
@@ -680,6 +693,16 @@ class SilhouettePlot(QGraphicsWidget):
                 self.setSelection(self.__selstate.selection)
             event.accept()
 
+    def keyPressEvent(self, event):
+        if event.key() in (Qt.Key_Up, Qt.Key_Down):
+            if event.key() == Qt.Key_Up:
+                self.__move_selection(self.selection(), -1)
+            elif event.key() == Qt.Key_Down:
+                self.__move_selection(self.selection(), 1)
+            event.accept()
+            return
+        super().keyPressEvent(event)
+
     def mouseMoveEvent(self, event):
         # Reimplemented
         if event.buttons() & Qt.LeftButton:
@@ -718,10 +741,17 @@ class SilhouettePlot(QGraphicsWidget):
                 rect = rect.adjusted(-1e-7, -1e-7, 1e-7, 1e-7)
 
             rect = rect.intersected(self.contentsRect())
-            action = action = self.__selstate.action & ~SelectAction.Current
+            action = self.__selstate.action & ~SelectAction.Current
             self.__setSelectionRect(rect, action)
             self.__selstate = None
 
+    def __move_selection(self, selection, offset):
+        ids = numpy.asarray([pi.data(0) for pi in self.__plotItems()]).ravel()
+        indices = [numpy.where(ids == i)[0] for i in selection]
+        indices = numpy.asarray(indices) + offset
+        if min(indices) >= 0 and max(indices) < len(ids):
+            self.setSelection(ids[indices])
+
     def __setSelectionRect(self, rect, action):
         # Set the current mouse drag selection rectangle
         if not rect.isValid():
@@ -819,13 +849,14 @@ class SilhouettePlot(QGraphicsWidget):
             items = item.items()
             if select.size:
                 for i in select:
-                    items[i].setBrush(Qt.red)
+                    color = numpy.hstack((grp.color, numpy.array([130])))
+                    items[i].setBrush(QBrush(QColor(*color)))
 
             deselect = numpy.flatnonzero(
                 numpy.in1d(grp.indices, deselected, assume_unique=True))
             if deselect.size:
                 for i in deselect:
-                    items[i].setBrush(self.__brush)
+                    items[i].setBrush(QBrush(QColor(*grp.color)))
 
     def __plotItems(self):
         for i in range(len(self.__groups)):
@@ -884,8 +915,7 @@ class BarPlotItem(QGraphicsWidget):
         return super().event(event)
 
     def sizeHint(self, which, constraint=QSizeF()):
-        spacing = max(self.__spacing * (self.count() - 1), 0)
-        return QSizeF(300, self.__barsize * self.count() + spacing)
+        return QSizeF(300, (self.__barsize + self.__spacing) * self.count())
 
     def setPreferredBarSize(self, size):
         if self.__barsize != size:
@@ -895,11 +925,6 @@ class BarPlotItem(QGraphicsWidget):
     def spacing(self):
         return self.__spacing
 
-    def setSpacing(self, spacing):
-        if self.__spacing != spacing:
-            self.__spacing = spacing
-            self.updateGeometry()
-
     def setPen(self, pen):
         pen = QPen(pen)
         if self.__pen != pen:
